# -*- coding: utf-8 -*-
# *****************************************************************************
# Copyright (c) 2016-2020, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# *****************************************************************************

import sys
import warnings

# import os
# myvar=os.environ

# import pprint
# pp = pprint.PrettyPrinter(indent=4)
# pp.pprint(myvar)

# os.system("ldd ./dpnp/libdpnp_backend_c.so")
# subprocess.call("echo Hello 123", shell=True)

if "numpy" in sys.modules:
    warnings.warn("\nDPNP: Module NumPy found. Please load DPNP module before NumPy.\n")

# from dpnp.dparray import dparray as ndarray
from dpnp.dpnp_array import dpnp_array as ndarray
from dpnp.dpnp_flatiter import flatiter as flatiter

from dpnp.dpnp_iface import *
from dpnp.dpnp_iface import __all__ as _iface__all__
from dpnp.dpnp_iface_types import *
from dpnp.version import __version__


__all__ = _iface__all__


dpnp_queue_initialize()
'''
SYCL queue initialization
'''



# cython: language_level=3
# -*- coding: utf-8 -*-
# *****************************************************************************
# Copyright (c) 2016-2020, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# *****************************************************************************

"""Example 10.

This example shows simple usage of the DPNP
in combination with dpCtl.

"""

import time

try:
    import dpnp
except ImportError:
    import os
    import sys

    sys.path.insert(0, os.path.abspath('.'))
    import dpnp

import numpy


def run(executor, size, test_type, repetition):
    x = executor.reshape(executor.arange(size * size, dtype=test_type), (size, size))

    times = []
    for _ in range(repetition):
        start_time = time.perf_counter()
        result = executor.sum(x)
        end_time = time.perf_counter()
        times.append(end_time - start_time)

    return numpy.median(times), result


def example():
    test_repetition = 5
    for test_type in [numpy.float64, numpy.float32, numpy.int64, numpy.int32]:
        type_name = numpy.dtype(test_type).name
        print(f"...Test data type is {type_name}, each test repetitions {test_repetition}")

        for size in [64, 128, 256, 512, 1024, 2048, 4096]:
            time_numpy, result_numpy = run(numpy, size, test_type, test_repetition)
            time_dpnp, result_dpnp = run(dpnp, size, test_type, test_repetition)

            if result_dpnp == result_numpy:
                verification = True
            else:
                verification = f"({result_dpnp} != {result_numpy})"

            msg = f"type:{type_name}:N:{size:4}:NumPy:{time_numpy:.3e}:SYCL:{time_dpnp:.3e}"
            msg += f":ratio:{time_numpy/time_dpnp:6.2f}:verification:{verification}"
            print(msg)


if __name__ == "__main__":
    try:
        import dpctl

        with dpctl.device_context("opencl:gpu") as gpu_queue:
            gpu_queue.get_sycl_device().dump_device_info()
            example()

    except ImportError:
        example()



import unittest

from tests.third_party.cupy import testing


class TestArithmetic(unittest.TestCase):

    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose()
    def test_modf_part1(self, xp, dtype):
        a = xp.array([-2.5, -1.5, -0.5, 0, 0.5, 1.5, 2.5], dtype=dtype)
        b, _ = xp.modf(a)

        return b

    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose()
    def test_modf_part2(self, xp, dtype):
        a = xp.array([-2.5, -1.5, -0.5, 0, 0.5, 1.5, 2.5], dtype=dtype)
        _, c = xp.modf(a)

        return c

    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose()
    def test_nanprod(self, xp, dtype):
        a = xp.array([-2.5, -1.5, xp.nan, 10.5, 1.5, xp.nan], dtype=dtype)
        return xp.nanprod(a)

    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose()
    def test_nansum(self, xp, dtype):
        a = xp.array([-2.5, -1.5, xp.nan, 10.5, 1.5, xp.nan], dtype=dtype)
        return xp.nansum(a)

    @testing.for_float_dtypes()
    @testing.numpy_cupy_allclose()
    def test_remainder(self, xp, dtype):
        a = xp.array([5, -3, -2, -1, -5], dtype=dtype)
        b = xp.full(a.size, 3, dtype=dtype)

        return xp.remainder(a, b)



import pytest

import dpnp

import numpy


class TestConvolve:
    def test_object(self):
        d = [1.] * 100
        k = [1.] * 3
        numpy.testing.assert_array_almost_equal(dpnp.convolve(d, k)[2:-2], dpnp.full(98, 3))

    def test_no_overwrite(self):
        d = dpnp.ones(100)
        k = dpnp.ones(3)
        dpnp.convolve(d, k)
        numpy.testing.assert_array_equal(d, dpnp.ones(100))
        numpy.testing.assert_array_equal(k, dpnp.ones(3))

    def test_mode(self):
        d = dpnp.ones(100)
        k = dpnp.ones(3)
        default_mode = dpnp.convolve(d, k, mode='full')
        full_mode = dpnp.convolve(d, k, mode='f')
        numpy.testing.assert_array_equal(full_mode, default_mode)
        # integer mode
        with numpy.testing.assert_raises(ValueError):
            dpnp.convolve(d, k, mode=-1)
        numpy.testing.assert_array_equal(dpnp.convolve(d, k, mode=2), full_mode)
        # illegal arguments
        with numpy.testing.assert_raises(TypeError):
            dpnp.convolve(d, k, mode=None)


@pytest.mark.parametrize("array",
                         [[[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]],
                          [[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]
                          ],
                         ids=['[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]]',
                              '[[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]'
                              ])
def test_diff(array):
    np_a = numpy.array(array)
    dpnp_a = dpnp.array(array)
    expected = numpy.diff(np_a)
    result = dpnp.diff(dpnp_a)
    numpy.testing.assert_allclose(expected, result)


@pytest.mark.parametrize("data",
                         [[[1 + 1j, -2j], [3 - 3j, 4j]]],
                         ids=['[[1+1j, -2j], [3-3j, 4j]]'])
def test_multiply_complex(data):
    np_a = numpy.array(data)
    dpnp_a = dpnp.array(data)

    result = dpnp.multiply(dpnp_a, dpnp_a)
    expected = numpy.multiply(np_a, np_a)
    numpy.testing.assert_array_equal(result, expected)

    result = dpnp.multiply(dpnp_a, 0.5j)
    expected = numpy.multiply(np_a, 0.5j)
    numpy.testing.assert_array_equal(result, expected)

    result = dpnp.multiply(0.5j, dpnp_a)
    expected = numpy.multiply(0.5j, np_a)
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("dtype1",
                         [numpy.bool_, numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.bool_', 'numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
@pytest.mark.parametrize("dtype2",
                         [numpy.bool_, numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.bool_', 'numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
@pytest.mark.parametrize("data",
                         [[[1, 2], [3, 4]]],
                         ids=['[[1, 2], [3, 4]]'])
def test_multiply_dtype(dtype1, dtype2, data):
    np_a = numpy.array(data, dtype=dtype1)
    dpnp_a = dpnp.array(data, dtype=dtype1)

    np_b = numpy.array(data, dtype=dtype2)
    dpnp_b = dpnp.array(data, dtype=dtype2)

    result = numpy.multiply(dpnp_a, dpnp_b)
    expected = numpy.multiply(np_a, np_b)
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("rhs", [[[1, 2, 3], [4, 5, 6]], [2.0, 1.5, 1.0], 3, 0.3])
@pytest.mark.parametrize("lhs", [[[6, 5, 4], [3, 2, 1]], [1.3, 2.6, 3.9], 5, 0.5])
@pytest.mark.parametrize("dtype", [numpy.int32, numpy.int64, numpy.float32, numpy.float64])
class TestMathematical:

    @staticmethod
    def array_or_scalar(xp, data, dtype=None):
        if numpy.isscalar(data):
            return data

        return xp.array(data, dtype=dtype)

    def _test_mathematical(self, name, dtype, lhs, rhs):
        a = self.array_or_scalar(dpnp, lhs, dtype=dtype)
        b = self.array_or_scalar(dpnp, rhs, dtype=dtype)
        result = getattr(dpnp, name)(a, b)

        a = self.array_or_scalar(numpy, lhs, dtype=dtype)
        b = self.array_or_scalar(numpy, rhs, dtype=dtype)
        expected = getattr(numpy, name)(a, b)

        numpy.testing.assert_allclose(result, expected, atol=1e-4)

    def test_add(self, dtype, lhs, rhs):
        self._test_mathematical('add', dtype, lhs, rhs)

    def test_arctan2(self, dtype, lhs, rhs):
        self._test_mathematical('arctan2', dtype, lhs, rhs)

    def test_copysign(self, dtype, lhs, rhs):
        self._test_mathematical('copysign', dtype, lhs, rhs)

    def test_divide(self, dtype, lhs, rhs):
        self._test_mathematical('divide', dtype, lhs, rhs)

    def test_fmod(self, dtype, lhs, rhs):
        self._test_mathematical('fmod', dtype, lhs, rhs)

    def test_floor_divide(self, dtype, lhs, rhs):
        self._test_mathematical('floor_divide', dtype, lhs, rhs)

    def test_hypot(self, dtype, lhs, rhs):
        self._test_mathematical('hypot', dtype, lhs, rhs)

    def test_maximum(self, dtype, lhs, rhs):
        self._test_mathematical('maximum', dtype, lhs, rhs)

    def test_minimum(self, dtype, lhs, rhs):
        self._test_mathematical('minimum', dtype, lhs, rhs)

    def test_multiply(self, dtype, lhs, rhs):
        self._test_mathematical('multiply', dtype, lhs, rhs)

    def test_remainder(self, dtype, lhs, rhs):
        self._test_mathematical('remainder', dtype, lhs, rhs)

    def test_power(self, dtype, lhs, rhs):
        self._test_mathematical('power', dtype, lhs, rhs)

    def test_subtract(self, dtype, lhs, rhs):
        self._test_mathematical('subtract', dtype, lhs, rhs)


@pytest.mark.parametrize("val_type",
                         [bool, int, float],
                         ids=['bool', 'int', 'float'])
@pytest.mark.parametrize("data_type",
                         [numpy.bool_, numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.bool_', 'numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
@pytest.mark.parametrize("val",
                         [0, 1, 5],
                         ids=['0', '1', '5'])
@pytest.mark.parametrize("array",
                         [[[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]],
                          [[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]],
                         ids=['[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]]',
                              '[[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]'])
def test_multiply_scalar(array, val, data_type, val_type):
    np_a = numpy.array(array, dtype=data_type)
    dpnp_a = dpnp.array(array, dtype=data_type)
    val_ = val_type(val)

    result = dpnp.multiply(np_a, val_)
    expected = numpy.multiply(dpnp_a, val_)
    numpy.testing.assert_array_equal(result, expected)

    result = dpnp.multiply(val_, np_a)
    expected = numpy.multiply(val_, dpnp_a)
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("shape",
                         [(), (3, 2)],
                         ids=['()', '(3, 2)'])
@pytest.mark.parametrize("dtype",
                         [numpy.float32, numpy.float64],
                         ids=['numpy.float32', 'numpy.float64'])
def test_multiply_scalar2(shape, dtype):
    np_a = numpy.ones(shape, dtype=dtype)
    dpnp_a = dpnp.ones(shape, dtype=dtype)

    result = 0.5 * dpnp_a
    expected = 0.5 * np_a
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("array", [[1, 2, 3, 4, 5],
                                   [1, 2, numpy.nan, 4, 5],
                                   [[1, 2, numpy.nan], [3, -4, -5]]])
def test_nancumprod(array):
    np_a = numpy.array(array)
    dpnp_a = dpnp.array(np_a)

    result = dpnp.nancumprod(dpnp_a)
    expected = numpy.nancumprod(np_a)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("array", [[1, 2, 3, 4, 5],
                                   [1, 2, numpy.nan, 4, 5],
                                   [[1, 2, numpy.nan], [3, -4, -5]]])
def test_nancumsum(array):
    np_a = numpy.array(array)
    dpnp_a = dpnp.array(np_a)

    result = dpnp.nancumsum(dpnp_a)
    expected = numpy.nancumsum(np_a)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("data",
                         [[[1., -1.], [0.1, -0.1]], [-2, -1, 0, 1, 2]],
                         ids=['[[1., -1.], [0.1, -0.1]]', '[-2, -1, 0, 1, 2]'])
@pytest.mark.parametrize("dtype",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
def test_negative(data, dtype):
    np_a = numpy.array(data, dtype=dtype)
    dpnp_a = dpnp.array(data, dtype=dtype)

    result = dpnp.negative(dpnp_a)
    expected = numpy.negative(np_a)
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("val_type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
@pytest.mark.parametrize("data_type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['numpy.float64', 'numpy.float32', 'numpy.int64', 'numpy.int32'])
@pytest.mark.parametrize("val",
                         [0, 1, 5],
                         ids=['0', '1', '5'])
@pytest.mark.parametrize("array",
                         [[[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]],
                          [[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]],
                         ids=['[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]]',
                              '[[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]'])
def test_power(array, val, data_type, val_type):
    np_a = numpy.array(array, dtype=data_type)
    dpnp_a = dpnp.array(array, dtype=data_type)
    val_ = val_type(val)
    result = dpnp.power(dpnp_a, val_)
    expected = numpy.power(np_a, val_)
    numpy.testing.assert_array_equal(expected, result)


class TestEdiff1d:
    @pytest.mark.parametrize("data_type",
                             [numpy.float64, numpy.float32, numpy.int64, numpy.int32])
    @pytest.mark.parametrize("array", [[1, 2, 4, 7, 0],
                                       [],
                                       [1],
                                       [[1, 2, 3], [5, 2, 8], [7, 3, 4]], ])
    def test_ediff1d_int(self, array, data_type):
        np_a = numpy.array(array, dtype=data_type)
        dpnp_a = dpnp.array(array, dtype=data_type)

        result = dpnp.ediff1d(dpnp_a)
        expected = numpy.ediff1d(np_a)
        numpy.testing.assert_array_equal(expected, result)

    def test_ediff1d_args(self):
        np_a = numpy.array([1, 2, 4, 7, 0])

        to_begin = numpy.array([-20, -30])
        to_end = numpy.array([20, 15])

        result = dpnp.ediff1d(np_a, to_end=to_end, to_begin=to_begin)
        expected = numpy.ediff1d(np_a, to_end=to_end, to_begin=to_begin)
        numpy.testing.assert_array_equal(expected, result)


class TestTrapz:
    @pytest.mark.parametrize("data_type",
                             [numpy.float64, numpy.float32, numpy.int64, numpy.int32])
    @pytest.mark.parametrize("array", [[1, 2, 3],
                                       [[1, 2, 3], [4, 5, 6]],
                                       [1, 4, 6, 9, 10, 12],
                                       [],
                                       [1]])
    def test_trapz_default(self, array, data_type):
        np_a = numpy.array(array, dtype=data_type)
        dpnp_a = dpnp.array(array, dtype=data_type)

        result = dpnp.trapz(dpnp_a)
        expected = numpy.trapz(np_a)
        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("data_type_y",
                             [numpy.float64, numpy.float32, numpy.int64, numpy.int32])
    @pytest.mark.parametrize("data_type_x",
                             [numpy.float64, numpy.float32, numpy.int64, numpy.int32])
    @pytest.mark.parametrize("y_array", [[1, 2, 4, 5],
                                         [1., 2.5, 6., 7.]])
    @pytest.mark.parametrize("x_array", [[2, 5, 6, 9]])
    def test_trapz_with_x_params(self, y_array, x_array, data_type_y, data_type_x):
        np_y = numpy.array(y_array, dtype=data_type_y)
        dpnp_y = dpnp.array(y_array, dtype=data_type_y)

        np_x = numpy.array(x_array, dtype=data_type_x)
        dpnp_x = dpnp.array(x_array, dtype=data_type_x)

        result = dpnp.trapz(dpnp_y, dpnp_x)
        expected = numpy.trapz(np_y, np_x)
        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("array", [[1, 2, 3], [4, 5, 6]])
    def test_trapz_with_x_param_2ndim(self, array):
        np_a = numpy.array(array)
        dpnp_a = dpnp.array(array)

        result = dpnp.trapz(dpnp_a, dpnp_a)
        expected = numpy.trapz(np_a, np_a)
        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("y_array", [[1, 2, 4, 5],
                                         [1., 2.5, 6., 7., ]])
    @pytest.mark.parametrize("dx", [2, 3, 4])
    def test_trapz_with_dx_params(self, y_array, dx):
        np_y = numpy.array(y_array)
        dpnp_y = dpnp.array(y_array)

        result = dpnp.trapz(dpnp_y, dx=dx)
        expected = numpy.trapz(np_y, dx=dx)
        numpy.testing.assert_array_equal(expected, result)


class TestCross:

    @pytest.mark.parametrize("axis", [None, 0],
                             ids=['None', '0'])
    @pytest.mark.parametrize("axisc", [-1, 0],
                             ids=['-1', '0'])
    @pytest.mark.parametrize("axisb", [-1, 0],
                             ids=['-1', '0'])
    @pytest.mark.parametrize("axisa", [-1, 0],
                             ids=['-1', '0'])
    @pytest.mark.parametrize("x1", [[1, 2, 3],
                                    [1., 2.5, 6.],
                                    [2, 4, 6]],
                             ids=['[1, 2, 3]',
                                  '[1., 2.5, 6.]',
                                  '[2, 4, 6]'])
    @pytest.mark.parametrize("x2", [[4, 5, 6],
                                    [1., 5., 2.],
                                    [6, 4, 3]],
                             ids=['[4, 5, 6]',
                                  '[1., 5., 2.]',
                                  '[6, 4, 3]'])
    def test_cross_3x3(self, x1, x2, axisa, axisb, axisc, axis):
        np_x1 = numpy.array(x1)
        dpnp_x1 = dpnp.array(x1)

        np_x2 = numpy.array(x2)
        dpnp_x2 = dpnp.array(x2)

        result = dpnp.cross(dpnp_x1, dpnp_x2, axisa, axisb, axisc, axis)
        expected = numpy.cross(np_x1, np_x2, axisa, axisb, axisc, axis)
        numpy.testing.assert_array_equal(expected, result)


class TestGradient:

    @pytest.mark.parametrize("array", [[2, 3, 6, 8, 4, 9],
                                       [3., 4., 7.5, 9.],
                                       [2, 6, 8, 10]])
    def test_gradient_y1(self, array):
        np_y = numpy.array(array)
        dpnp_y = dpnp.array(array)

        result = dpnp.gradient(dpnp_y)
        expected = numpy.gradient(np_y)
        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("array", [[2, 3, 6, 8, 4, 9],
                                       [3., 4., 7.5, 9.],
                                       [2, 6, 8, 10]])
    @pytest.mark.parametrize("dx", [2, 3.5])
    def test_gradient_y1_dx(self, array, dx):
        np_y = numpy.array(array)
        dpnp_y = dpnp.array(array)

        result = dpnp.gradient(dpnp_y, dx)
        expected = numpy.gradient(np_y, dx)
        numpy.testing.assert_array_equal(expected, result)


class TestGradient:

    @pytest.mark.parametrize("array", [[2, 3, 6, 8, 4, 9],
                                       [3., 4., 7.5, 9.],
                                       [2, 6, 8, 10]])
    def test_gradient_y1(self, array):
        np_y = numpy.array(array)
        dpnp_y = dpnp.array(array)

        result = dpnp.gradient(dpnp_y)
        expected = numpy.gradient(np_y)
        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("array", [[2, 3, 6, 8, 4, 9],
                                       [3., 4., 7.5, 9.],
                                       [2, 6, 8, 10]])
    @pytest.mark.parametrize("dx", [2, 3.5])
    def test_gradient_y1_dx(self, array, dx):
        np_y = numpy.array(array)
        dpnp_y = dpnp.array(array)

        result = dpnp.gradient(dpnp_y, dx)
        expected = numpy.gradient(np_y, dx)
        numpy.testing.assert_array_equal(expected, result)


class TestCeil:

    def test_ceil(self):
        array_data = numpy.arange(10)
        out = numpy.empty(10, dtype=numpy.float64)

        # DPNP
        dp_array = dpnp.array(array_data, dtype=dpnp.float64)
        dp_out = dpnp.array(out, dtype=dpnp.float64)
        result = dpnp.ceil(dp_array, out=dp_out)

        # original
        np_array = numpy.array(array_data, dtype=numpy.float64)
        expected = numpy.ceil(np_array, out=out)

        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("dtype",
                             [numpy.float32, numpy.int64, numpy.int32],
                             ids=['numpy.float32', 'numpy.int64', 'numpy.int32'])
    def test_invalid_dtype(self, dtype):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(10, dtype=dtype)

        with pytest.raises(ValueError):
            dpnp.ceil(dp_array, out=dp_out)

    @pytest.mark.parametrize("shape",
                             [(0,), (15, ), (2, 2)],
                             ids=['(0,)', '(15, )', '(2,2)'])
    def test_invalid_shape(self, shape):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(shape, dtype=dpnp.float64)

        with pytest.raises(ValueError):
            dpnp.ceil(dp_array, out=dp_out)


class TestFloor:

    def test_floor(self):
        array_data = numpy.arange(10)
        out = numpy.empty(10, dtype=numpy.float64)

        # DPNP
        dp_array = dpnp.array(array_data, dtype=dpnp.float64)
        dp_out = dpnp.array(out, dtype=dpnp.float64)
        result = dpnp.floor(dp_array, out=dp_out)

        # original
        np_array = numpy.array(array_data, dtype=numpy.float64)
        expected = numpy.floor(np_array, out=out)

        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("dtype",
                             [numpy.float32, numpy.int64, numpy.int32],
                             ids=['numpy.float32', 'numpy.int64', 'numpy.int32'])
    def test_invalid_dtype(self, dtype):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(10, dtype=dtype)

        with pytest.raises(ValueError):
            dpnp.floor(dp_array, out=dp_out)

    @pytest.mark.parametrize("shape",
                             [(0,), (15, ), (2, 2)],
                             ids=['(0,)', '(15, )', '(2,2)'])
    def test_invalid_shape(self, shape):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(shape, dtype=dpnp.float64)

        with pytest.raises(ValueError):
            dpnp.floor(dp_array, out=dp_out)


class TestTrunc:

    def test_trunc(self):
        array_data = numpy.arange(10)
        out = numpy.empty(10, dtype=numpy.float64)

        # DPNP
        dp_array = dpnp.array(array_data, dtype=dpnp.float64)
        dp_out = dpnp.array(out, dtype=dpnp.float64)
        result = dpnp.trunc(dp_array, out=dp_out)

        # original
        np_array = numpy.array(array_data, dtype=numpy.float64)
        expected = numpy.trunc(np_array, out=out)

        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("dtype",
                             [numpy.float32, numpy.int64, numpy.int32],
                             ids=['numpy.float32', 'numpy.int64', 'numpy.int32'])
    def test_invalid_dtype(self, dtype):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(10, dtype=dtype)

        with pytest.raises(ValueError):
            dpnp.trunc(dp_array, out=dp_out)

    @pytest.mark.parametrize("shape",
                             [(0,), (15, ), (2, 2)],
                             ids=['(0,)', '(15, )', '(2,2)'])
    def test_invalid_shape(self, shape):

        dp_array = dpnp.arange(10, dtype=dpnp.float64)
        dp_out = dpnp.empty(shape, dtype=dpnp.float64)

        with pytest.raises(ValueError):
            dpnp.trunc(dp_array, out=dp_out)


class TestPower:

    def test_power(self):
        array1_data = numpy.arange(10)
        array2_data = numpy.arange(5, 15)
        out = numpy.empty(10, dtype=numpy.float64)

        # DPNP
        dp_array1 = dpnp.array(array1_data, dtype=dpnp.float64)
        dp_array2 = dpnp.array(array2_data, dtype=dpnp.float64)
        dp_out = dpnp.array(out, dtype=dpnp.float64)
        result = dpnp.power(dp_array1, dp_array2, out=dp_out)

        # original
        np_array1 = numpy.array(array1_data, dtype=numpy.float64)
        np_array2 = numpy.array(array2_data, dtype=numpy.float64)
        expected = numpy.power(np_array1, np_array2, out=out)

        numpy.testing.assert_array_equal(expected, result)

    @pytest.mark.parametrize("dtype",
                             [numpy.float32, numpy.int64, numpy.int32],
                             ids=['numpy.float32', 'numpy.int64', 'numpy.int32'])
    def test_invalid_dtype(self, dtype):

        dp_array1 = dpnp.arange(10, dtype=dpnp.float64)
        dp_array2 = dpnp.arange(5, 15, dtype=dpnp.float64)
        dp_out = dpnp.empty(10, dtype=dtype)

        with pytest.raises(ValueError):
            dpnp.power(dp_array1, dp_array2, out=dp_out)

    @pytest.mark.parametrize("shape",
                             [(0,), (15, ), (2, 2)],
                             ids=['(0,)', '(15, )', '(2,2)'])
    def test_invalid_shape(self, shape):

        dp_array1 = dpnp.arange(10, dtype=dpnp.float64)
        dp_array2 = dpnp.arange(5, 15, dtype=dpnp.float64)
        dp_out = dpnp.empty(shape, dtype=dpnp.float64)

        with pytest.raises(ValueError):
            dpnp.power(dp_array1, dp_array2, out=dp_out)



# cython: language_level=3
# distutils: language = c++
# -*- coding: utf-8 -*-
# *****************************************************************************
# Copyright (c) 2016-2020, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# *****************************************************************************

"""
Interface of the sorting function of the dpnp

Notes
-----
This module is a face or public interface file for the library
it contains:
 - Interface functions
 - documentation for the functions
 - The functions parameters check

"""


import numpy

from dpnp.dpnp_algo import *
from dpnp.dpnp_utils import *

import dpnp


__all__ = [
    'argsort',
    'partition',
    'searchsorted',
    'sort'
]


def argsort(in_array1, axis=-1, kind=None, order=None):
    """
    Returns the indices that would sort an array.

    For full documentation refer to :obj:`numpy.argsort`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Otherwise the function will be executed sequentially on CPU.
    Prameters ``axis`` is supported only with default value ``-1``.
    Prameters ``kind`` is supported only with default value ``None``.
    Prameters ``order`` is supported only with default value ``None``.
    Input array data types are limited by supported DPNP :ref:`Data types`.

    See Also
    --------
    :obj:`dpnp.sort` : Describes sorting algorithms used.
    :obj:`dpnp.lexsort` : Indirect stable sort with multiple keys.
    :obj:`dpnp.argpartition` : Indirect partial sort.
    :obj:`dpnp.take_along_axis` : Apply ``index_array`` from argsort to
                                  an array as if by calling sort.

    Examples
    --------
    >>> import dpnp as np
    >>> x = np.array([3, 1, 2])
    >>> out = np.argsort(x)
    >>> [i for i in out]
    [1, 2, 0]

    """

    x1_desc = dpnp.get_dpnp_descriptor(in_array1)
    if x1_desc:
        if axis != -1:
            pass
        elif kind is not None:
            pass
        elif order is not None:
            pass
        else:
            return dpnp_argsort(x1_desc).get_pyobj()

    return call_origin(numpy.argsort, in_array1, axis, kind, order)


def partition(x1, kth, axis=-1, kind='introselect', order=None):
    """
    Return a partitioned copy of an array.
    For full documentation refer to :obj:`numpy.partition`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Input kth is supported as :obj:`int`.
    Parameters ``axis``, ``kind`` and ``order`` are supported only with default values.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        if not isinstance(kth, int):
            pass
        elif x1_desc.ndim == 0:
            pass
        elif kth >= x1_desc.shape[x1_desc.ndim - 1] or x1_desc.ndim + kth < 0:
            pass
        elif axis != -1:
            pass
        elif kind != 'introselect':
            pass
        elif order is not None:
            pass
        else:
            return dpnp_partition(x1_desc, kth, axis, kind, order).get_pyobj()

    return call_origin(numpy.partition, x1, kth, axis, kind, order)


def searchsorted(x1, x2, side='left', sorter=None):
    """
    Find indices where elements should be inserted to maintain order.
    For full documentation refer to :obj:`numpy.searchsorted`.

    Limitations
    -----------
    Input arrays is supported as :obj:`dpnp.ndarray`.
    Input array is supported only sorted.
    Input side is supported only values ``left``, ``right``.
    Parameters ``sorter`` is supported only with default values.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    x2_desc = dpnp.get_dpnp_descriptor(x2)
    if 0 and x1_desc and x2_desc:
        if x1_desc.ndim != 1:
            pass
        elif x1_desc.dtype != x2_desc.dtype:
            pass
        elif side not in ['left', 'right']:
            pass
        elif sorter is not None:
            pass
        elif x1_desc.size < 2:
            pass
        else:
            return dpnp_searchsorted(x1_desc, x2_desc, side=side).get_pyobj()

    return call_origin(numpy.searchsorted, x1, x2, side=side, sorter=sorter)


def sort(x1, **kwargs):
    """
    Return a sorted copy of an array.

    For full documentation refer to :obj:`numpy.sort`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Keyword arguments ``kwargs`` are currently unsupported.
    Dimension of input array is supported to be equal to ``1``.
    Otherwise the function will be executed sequentially on CPU.
    Input array data types are limited by supported DPNP :ref:`Data types`.

    See Also
    --------
    :obj:`dpnp.argsort` : Indirect sort.
    :obj:`dpnp.lexsort` : Indirect stable sort on multiple keys.
    :obj:`dpnp.searchsorted` : Find elements in a sorted array.
    :obj:`dpnp.partition` : Partial sort.

    Examples
    --------
    >>> import dpnp as np
    >>> a = np.array([1, 4, 3, 1])
    >>> out = np.sort(a)
    >>> [i for i in out]
    [1, 1, 3, 4]

    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc and not kwargs:
        if x1_desc.ndim != 1:
            pass
        else:
            return dpnp_sort(x1_desc).get_pyobj()

    return call_origin(numpy.sort, x1, **kwargs)



import pytest

import dpnp

import numpy


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize("size",
                         [2, 4, 8, 16, 3, 9, 27, 81])
def test_count_nonzero(type, size):
    a = numpy.arange(size, dtype=type)
    for i in range(int(size / 2)):
        a[(i * (int(size / 3) - 1)) % size] = 0

    ia = dpnp.array(a)

    np_res = numpy.count_nonzero(a)
    dpnp_res = dpnp.count_nonzero(ia)

    numpy.testing.assert_allclose(dpnp_res, np_res)



import time


def cos_2_args(executor, size, test_type):
    """ sin(A + B) = sin A cos B + cos A sin B """

    start_time = time.perf_counter()
    input_A = executor.arange(size, dtype=test_type)
    input_B = executor.arange(size, dtype=test_type)
    end_time = time.perf_counter()
    memalloc_time = end_time - start_time

    start_time = time.perf_counter()

    sin_A = executor.sin(input_A)
    cos_B = executor.cos(input_B)
    sincosA = sin_A * cos_B

    cos_A = executor.cos(input_A)
    sin_B = executor.sin(input_B)
    sincosB = cos_A * sin_B

    result = sincosA + sincosB

    end_time = time.perf_counter()
    calculation_time = end_time - start_time

    print(f"memalloc_time={memalloc_time}, calculation_time={calculation_time}, executor={executor}")

    return result


if __name__ == '__main__':
    size = 33554432  # 16777216

    import dpnp
    cos_2_args(dpnp, size, dpnp.float64)

    import numpy
    cos_2_args(numpy, size, numpy.float64)



import pytest

import dpnp
import numpy


@pytest.mark.parametrize("dtype",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=["float64", "float32", "int64", "int32"])
@pytest.mark.parametrize("data",
                         [[1, 2, 3], [1., 2., 3.]],
                         ids=["[1, 2, 3]", "[1., 2., 3.]"])
def test_asfarray(dtype, data):
    expected = numpy.asfarray(data, dtype)
    result = dpnp.asfarray(data, dtype)

    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("dtype",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=["float64", "float32", "int64", "int32"])
@pytest.mark.parametrize("data",
                         [[1, 2, 3], [1., 2., 3.]],
                         ids=["[1, 2, 3]", "[1., 2., 3.]"])
def test_asfarray2(dtype, data):
    expected = numpy.asfarray(numpy.array(data), dtype)
    result = dpnp.asfarray(dpnp.array(data), dtype)

    numpy.testing.assert_array_equal(result, expected)


class TestConcatenate:
    def test_returns_copy(self):
        a = dpnp.array(numpy.eye(3))
        b = dpnp.concatenate([a])
        b[0, 0] = 2
        assert b[0, 0] != a[0, 0]

    def test_large_concatenate_axis_None(self):
        x = dpnp.arange(1, 100)
        r = dpnp.concatenate(x, None)
        numpy.testing.assert_array_equal(x, r)
        r = dpnp.concatenate(x, 100)
        numpy.testing.assert_array_equal(x, r)

    def test_concatenate(self):
        # Test concatenate function
        # One sequence returns unmodified (but as array)
        r4 = list(range(4))
        numpy.testing.assert_array_equal(dpnp.concatenate((r4,)), r4)
        # Any sequence
        numpy.testing.assert_array_equal(dpnp.concatenate((tuple(r4),)), r4)
        numpy.testing.assert_array_equal(dpnp.concatenate((dpnp.array(r4),)), r4)
        # 1D default concatenation
        r3 = list(range(3))
        numpy.testing.assert_array_equal(dpnp.concatenate((r4, r3)), r4 + r3)
        # Mixed sequence types
        numpy.testing.assert_array_equal(dpnp.concatenate((tuple(r4), r3)), r4 + r3)
        numpy.testing.assert_array_equal(dpnp.concatenate((dpnp.array(r4), r3)), r4 + r3)
        # Explicit axis specification
        numpy.testing.assert_array_equal(dpnp.concatenate((r4, r3), 0), r4 + r3)
        # Including negative
        numpy.testing.assert_array_equal(dpnp.concatenate((r4, r3), -1), r4 + r3)
        # 2D
        a23 = dpnp.array([[10, 11, 12], [13, 14, 15]])
        a13 = dpnp.array([[0, 1, 2]])
        res = dpnp.array([[10, 11, 12], [13, 14, 15], [0, 1, 2]])
        numpy.testing.assert_array_equal(dpnp.concatenate((a23, a13)), res)
        numpy.testing.assert_array_equal(dpnp.concatenate((a23, a13), 0), res)
        numpy.testing.assert_array_equal(dpnp.concatenate((a23.T, a13.T), 1), res.T)
        numpy.testing.assert_array_equal(dpnp.concatenate((a23.T, a13.T), -1), res.T)
        # Arrays much match shape
        numpy.testing.assert_raises(ValueError, dpnp.concatenate, (a23.T, a13.T), 0)
        # 3D
        res = dpnp.reshape(dpnp.arange(2 * 3 * 7), (2, 3, 7))
        a0 = res[..., :4]
        a1 = res[..., 4:6]
        a2 = res[..., 6:]
        numpy.testing.assert_array_equal(dpnp.concatenate((a0, a1, a2), 2), res)
        numpy.testing.assert_array_equal(dpnp.concatenate((a0, a1, a2), -1), res)
        numpy.testing.assert_array_equal(dpnp.concatenate((a0.T, a1.T, a2.T), 0), res.T)

        out = dpnp.copy(res)
        rout = dpnp.concatenate((a0, a1, a2), 2, out=out)
        numpy.testing.assert_(out is rout)
        numpy.testing.assert_equal(res, rout)


class TestHstack:
    def test_non_iterable(self):
        numpy.testing.assert_raises(TypeError, dpnp.hstack, 1)

    def test_empty_input(self):
        numpy.testing.assert_raises(ValueError, dpnp.hstack, ())

    def test_0D_array(self):
        b = dpnp.array(2)
        a = dpnp.array(1)
        res = dpnp.hstack([a, b])
        desired = dpnp.array([1, 2])
        numpy.testing.assert_array_equal(res, desired)

    def test_1D_array(self):
        a = dpnp.array([1])
        b = dpnp.array([2])
        res = dpnp.hstack([a, b])
        desired = dpnp.array([1, 2])
        numpy.testing.assert_array_equal(res, desired)

    def test_2D_array(self):
        a = dpnp.array([[1], [2]])
        b = dpnp.array([[1], [2]])
        res = dpnp.hstack([a, b])
        desired = dpnp.array([[1, 1], [2, 2]])
        numpy.testing.assert_array_equal(res, desired)

    def test_generator(self):
        with numpy.testing.assert_warns(FutureWarning):
            dpnp.hstack((numpy.arange(3) for _ in range(2)))
        with numpy.testing.assert_warns(FutureWarning):
            dpnp.hstack(map(lambda x: x, numpy.ones((3, 2))))


class TestVstack:
    def test_non_iterable(self):
        numpy.testing.assert_raises(TypeError, dpnp.vstack, 1)

    def test_empty_input(self):
        numpy.testing.assert_raises(ValueError, dpnp.vstack, ())

    def test_0D_array(self):
        a = dpnp.array(1)
        b = dpnp.array(2)
        res = dpnp.vstack([a, b])
        desired = dpnp.array([[1], [2]])
        numpy.testing.assert_array_equal(res, desired)

    def test_1D_array(self):
        a = dpnp.array([1])
        b = dpnp.array([2])
        res = dpnp.vstack([a, b])
        desired = dpnp.array([[1], [2]])
        numpy.testing.assert_array_equal(res, desired)

    def test_2D_array(self):
        a = dpnp.array([[1], [2]])
        b = dpnp.array([[1], [2]])
        res = dpnp.vstack([a, b])
        desired = dpnp.array([[1], [2], [1], [2]])
        numpy.testing.assert_array_equal(res, desired)

    def test_2D_array2(self):
        a = dpnp.array([1, 2])
        b = dpnp.array([1, 2])
        res = dpnp.vstack([a, b])
        desired = dpnp.array([[1, 2], [1, 2]])
        numpy.testing.assert_array_equal(res, desired)

    def test_generator(self):
        with numpy.testing.assert_warns(FutureWarning):
            dpnp.vstack((numpy.arange(3) for _ in range(2)))



import pytest
import numpy
import dpnp


testdata = []
testdata += [([True, False, True], dtype) for dtype in ['float32', 'float64', 'int32', 'int64', 'bool']]
testdata += [([1, -1, 0], dtype) for dtype in ['float32', 'float64', 'int32', 'int64']]
testdata += [([0.1, 0.0, -0.1], dtype) for dtype in ['float32', 'float64']]
testdata += [([1j, -1j, 1 - 2j], dtype) for dtype in ['complex128']]


@pytest.mark.parametrize("in_obj,out_dtype", testdata)
def test_copyto_dtype(in_obj, out_dtype):
    ndarr = numpy.array(in_obj)
    expected = numpy.empty(ndarr.size, dtype=out_dtype)
    numpy.copyto(expected, ndarr)

    dparr = dpnp.array(in_obj)
    result = dpnp.empty(dparr.size, dtype=out_dtype)
    dpnp.copyto(result, dparr)

    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("arr",
                         [[], [1, 2, 3, 4], [[1, 2], [3, 4]], [[[1], [2]], [[3], [4]]]],
                         ids=['[]', '[1, 2, 3, 4]', '[[1, 2], [3, 4]]', '[[[1], [2]], [[3], [4]]]'])
def test_repeat(arr):
    a = numpy.array(arr)
    dpnp_a = dpnp.array(arr)
    expected = numpy.repeat(a, 2)
    result = dpnp.repeat(dpnp_a, 2)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("array",
                         [[1, 2, 3],
                          [1, 2, 2, 1, 2, 4],
                          [2, 2, 2, 2],
                          []],
                         ids=['[1, 2, 3]',
                              '[1, 2, 2, 1, 2, 4]',
                              '[2, 2, 2, 2]',
                              '[]'])
def test_unique(array):
    np_a = numpy.array(array)
    dpnp_a = dpnp.array(array)

    expected = numpy.unique(np_a)
    result = dpnp.unique(dpnp_a)
    numpy.testing.assert_array_equal(expected, result)



import pytest

import dpnp

import numpy

import tempfile


@pytest.mark.parametrize("k",
                         [-6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6],
                         ids=['-6', '-5', '-4', '-3', '-2', '-1', '0', '1', '2', '3', '4', '5', '6'])
@pytest.mark.parametrize("v",
                         [[0, 1, 2, 3, 4],
                          [1, 1, 1, 1, 1],
                          [[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                          [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]],
                         ids=['[0, 1, 2, 3, 4]',
                              '[1, 1, 1, 1, 1]',
                              '[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[0, 1, 2], [3, 4, 5], [6, 7, 8]]',
                              '[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]'])
def test_diag(v, k):
    a = numpy.array(v)
    ia = dpnp.array(a)
    expected = numpy.diag(a, k)
    result = dpnp.diag(ia, k)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_frombuffer(type):
    buffer = b'12345678'

    np_res = numpy.frombuffer(buffer, dtype=type)
    dpnp_res = dpnp.frombuffer(buffer, dtype=type)

    numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_fromfile(type):
    with tempfile.TemporaryFile() as fh:
        fh.write(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08")
        fh.flush()

        fh.seek(0)
        np_res = numpy.fromfile(fh, dtype=type)
        fh.seek(0)
        dpnp_res = dpnp.fromfile(fh, dtype=type)

        numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_fromfunction(type):
    def func(x, y):
        return x * y

    shape = (3, 3)

    np_res = numpy.fromfunction(func, shape=shape, dtype=type)
    dpnp_res = dpnp.fromfunction(func, shape=shape, dtype=type)

    numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_fromiter(type):
    iter = [1, 2, 3, 4]

    np_res = numpy.fromiter(iter, dtype=type)
    dpnp_res = dpnp.fromiter(iter, dtype=type)

    numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_fromstring(type):
    string = "1 2 3 4"

    np_res = numpy.fromstring(string, dtype=type, sep=' ')
    dpnp_res = dpnp.fromstring(string, dtype=type, sep=' ')

    numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize("num",
                         [2, 4, 8, 3, 9, 27])
@pytest.mark.parametrize("endpoint",
                         [True, False])
def test_geomspace(type, num, endpoint):
    start = 2
    stop = 256

    np_res = numpy.geomspace(start, stop, num, endpoint, type)
    dpnp_res = dpnp.geomspace(start, stop, num, endpoint, type)

    # Note that the above may not produce exact integers:
    # (c) https://numpy.org/doc/stable/reference/generated/numpy.geomspace.html
    if type in [numpy.int64, numpy.int32]:
        numpy.testing.assert_allclose(dpnp_res, np_res, atol=1)
    else:
        numpy.testing.assert_allclose(dpnp_res, np_res)


@pytest.mark.parametrize("n",
                         [0, 1, 4],
                         ids=['0', '1', '4'])
@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64,
                          numpy.int32, numpy.bool, numpy.complex128, None],
                         ids=['float64', 'float32', 'int64', 'int32', 'bool', 'complex128', 'None'])
def test_identity(n, type):
    expected = numpy.identity(n, dtype=type)
    result = dpnp.identity(n, dtype=type)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
def test_loadtxt(type):
    with tempfile.TemporaryFile() as fh:
        fh.write(b"1 2 3 4")
        fh.flush()

        fh.seek(0)
        np_res = numpy.loadtxt(fh, dtype=type)
        fh.seek(0)
        dpnp_res = dpnp.loadtxt(fh, dtype=type)

        numpy.testing.assert_array_equal(dpnp_res, np_res)


@pytest.mark.parametrize("dtype",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64, numpy.int32],
                         ids=['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize("offset",
                         [0, 1],
                         ids=['0', '1'])
@pytest.mark.parametrize("array",
                         [[[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                          [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
                          [[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]],
                          [[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]],
                          [[[[1, 2, 3], [3, 4, 5]], [[1, 2, 3], [2, 1, 0]]], [
                              [[1, 3, 5], [3, 1, 0]], [[0, 1, 2], [1, 3, 4]]]],
                          [[[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]], [[[13, 14, 15], [16, 17, 18]], [[19, 20, 21], [22, 23, 24]]]]],
                         ids=['[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[0, 1, 2], [3, 4, 5], [6, 7, 8]]',
                              '[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]',
                              '[[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]]',
                              '[[[[1, 2], [3, 4]], [[1, 2], [2, 1]]], [[[1, 3], [3, 1]], [[0, 1], [1, 3]]]]',
                              '[[[[1, 2, 3], [3, 4, 5]], [[1, 2, 3], [2, 1, 0]]], [[[1, 3, 5], [3, 1, 0]], [[0, 1, 2], [1, 3, 4]]]]',
                              '[[[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]], [[[13, 14, 15], [16, 17, 18]], [[19, 20, 21], [22, 23, 24]]]]'])
def test_trace(array, offset, type, dtype):
    a = numpy.array(array, type)
    ia = dpnp.array(array, type)
    expected = numpy.trace(a, offset=offset, dtype=dtype)
    result = dpnp.trace(ia, offset=offset, dtype=dtype)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("N",
                         [0, 1, 2, 3, 4],
                         ids=['0', '1', '2', '3', '4'])
@pytest.mark.parametrize("M",
                         [0, 1, 2, 3, 4],
                         ids=['0', '1', '2', '3', '4'])
@pytest.mark.parametrize("k",
                         [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5],
                         ids=['-5', '-4', '-3', '-2', '-1', '0', '1', '2', '3', '4', '5'])
@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, float, numpy.int64, numpy.int32, numpy.int, numpy.float, int],
                         ids=['float64', 'float32', 'numpy.float', 'float', 'int64', 'int32', 'numpy.int', 'int'])
def test_tri(N, M, k, type):
    expected = numpy.tri(N, M, k, dtype=type)
    result = dpnp.tri(N, M, k, dtype=type)
    numpy.testing.assert_array_equal(result, expected)


def test_tri_default_dtype():
    expected = numpy.tri(3, 5, -1)
    result = dpnp.tri(3, 5, -1)
    numpy.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("k",
                         [-6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6],
                         ids=['-6', '-5', '-4', '-3', '-2', '-1', '0', '1', '2', '3', '4', '5', '6'])
@pytest.mark.parametrize("m",
                         [[0, 1, 2, 3, 4],
                          [1, 1, 1, 1, 1],
                          [[0, 0], [0, 0]],
                          [[1, 2], [1, 2]],
                          [[1, 2], [3, 4]],
                          [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                          [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]],
                         ids=['[0, 1, 2, 3, 4]',
                              '[1, 1, 1, 1, 1]',
                              '[[0, 0], [0, 0]]',
                              '[[1, 2], [1, 2]]',
                              '[[1, 2], [3, 4]]',
                              '[[0, 1, 2], [3, 4, 5], [6, 7, 8]]',
                              '[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]'])
def test_tril(m, k):
    a = numpy.array(m)
    ia = dpnp.array(a)
    expected = numpy.tril(a, k)
    result = dpnp.tril(ia, k)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("k",
                         [-4, -3, -2, -1, 0, 1, 2, 3, 4],
                         ids=['-4', '-3', '-2', '-1', '0', '1', '2', '3', '4'])
@pytest.mark.parametrize("m",
                         [[0, 1, 2, 3, 4],
                          [[1, 2], [3, 4]],
                          [[0, 1, 2], [3, 4, 5], [6, 7, 8]],
                          [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]],
                         ids=['[0, 1, 2, 3, 4]',
                              '[[1, 2], [3, 4]]',
                              '[[0, 1, 2], [3, 4, 5], [6, 7, 8]]',
                              '[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]'])
def test_triu(m, k):
    a = numpy.array(m)
    ia = dpnp.array(a)
    expected = numpy.triu(a, k)
    result = dpnp.triu(ia, k)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("k",
                         [-4, -3, -2, -1, 0, 1, 2, 3, 4],
                         ids=['-4', '-3', '-2', '-1', '0', '1', '2', '3', '4'])
def test_triu_size_null(k):
    a = numpy.ones(shape=(1, 2, 0))
    ia = dpnp.array(a)
    expected = numpy.triu(a, k)
    result = dpnp.triu(ia, k)
    numpy.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize("array",
                         [[1, 2, 3, 4],
                          [],
                          [0, 3, 5]],
                         ids=['[1, 2, 3, 4]',
                              '[]',
                              '[0, 3, 5]'])
@pytest.mark.parametrize("type",
                         [numpy.float64, numpy.float32, numpy.int64,
                          numpy.int32, numpy.bool, numpy.complex128],
                         ids=['float64', 'float32', 'int64', 'int32', 'bool', 'complex128'])
@pytest.mark.parametrize("n",
                         [0, 1, 4, None],
                         ids=['0', '1', '4', 'None'])
@pytest.mark.parametrize("increase",
                         [True, False],
                         ids=['True', 'False'])
def test_vander(array, type, n, increase):
    a_np = numpy.array(array, dtype=type)
    a_dpnp = dpnp.array(array, dtype=type)

    expected = numpy.vander(a_np, N=n, increasing=increase)
    result = dpnp.vander(a_dpnp, N=n, increasing=increase)
    numpy.testing.assert_array_equal(expected, result)



import dpnp
import numpy
import pytest

from tests.tests_perf.data_generator import *
from tests.tests_perf.test_perf_base import DPNPTestPerfBase


class TestDPNP(DPNPTestPerfBase):

    @pytest.mark.parametrize("dtype", [numpy.float64, numpy.float32, numpy.int64, numpy.int32])
    @pytest.mark.parametrize("size", [32, 64, 128, 256])  # , 512, 1024, 2048, 4096])
    def test_matmul(self, lib, dtype, size):
        input1 = gen_array_2d(lib, size, size, dtype=dtype, seed=self.seed)
        input2 = gen_array_2d(lib, size, size, dtype=dtype, seed=self.seed)

        self.dpnp_benchmark("matmul", lib, dtype, input1.size, input1, input2)



# cython: language_level=3
# distutils: language = c++
# -*- coding: utf-8 -*-
# *****************************************************************************
# Copyright (c) 2016-2020, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# *****************************************************************************

"""
Interface of the Indexing part of the DPNP

Notes
-----
This module is a face or public interface file for the library
it contains:
 - Interface functions
 - documentation for the functions
 - The functions parameters check

"""


import collections

from dpnp.dpnp_algo import *
from dpnp.dpnp_utils import *

import dpnp
import numpy


__all__ = [
    "choose",
    "diag_indices",
    "diag_indices_from",
    "diagonal",
    "fill_diagonal",
    "indices",
    "nonzero",
    "place",
    "put",
    "put_along_axis",
    "putmask",
    "select",
    "take",
    "take_along_axis",
    "tril_indices",
    "tril_indices_from",
    "triu_indices",
    "triu_indices_from"
]


def choose(x1, choices, out=None, mode='raise'):
    """
    Construct an array from an index array and a set of arrays to choose from.

    For full documentation refer to :obj:`numpy.choose`.

    See also
    --------
    :obj:`take_along_axis` : Preferable if choices is an array.
    """
    x1_desc = dpnp.get_dpnp_descriptor(x1)

    choices_list = []
    for choice in choices:
        choices_list.append(dpnp.get_dpnp_descriptor(choice))

    if x1_desc:
        if any(not desc for desc in choices_list):
            pass
        elif out is not None:
            pass
        elif mode != 'raise':
            pass
        elif any(not choices[0].dtype == choice.dtype for choice in choices):
            pass
        elif not len(choices_list):
            pass
        else:
            size = x1_desc.size
            choices_size = choices_list[0].size
            if any(choice.size != choices_size or choice.size != size for choice in choices):
                pass
            elif any(x >= choices_size for x in dpnp.asnumpy(x1)):
                pass
            else:
                return dpnp_choose(x1_desc, choices_list).get_pyobj()

    return call_origin(numpy.choose, x1, choices, out, mode)


def diag_indices(n, ndim=2):
    """
    Return the indices to access the main diagonal of an array.

    This returns a tuple of indices that can be used to access the main
    diagonal of an array `a` with ``a.ndim >= 2`` dimensions and shape
    (n, n, ..., n). For ``a.ndim = 2`` this is the usual diagonal, for
    ``a.ndim > 2`` this is the set of indices to access ``a[i, i, ..., i]``
    for ``i = [0..n-1]``.

    For full documentation refer to :obj:`numpy.diag_indices`.

    See also
    --------
    :obj:`diag_indices_from` : Return the indices to access the main
                               diagonal of an n-dimensional array.

    Examples
    --------
    Create a set of indices to access the diagonal of a (4, 4) array:

    >>> import dpnp as np
    >>> di = np.diag_indices(4)
    >>> di
    (array([0, 1, 2, 3]), array([0, 1, 2, 3]))
    >>> a = np.arange(16).reshape(4, 4)
    >>> a
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11],
           [12, 13, 14, 15]])
    >>> a[di] = 100
    >>> a
    array([[100,   1,   2,   3],
           [  4, 100,   6,   7],
           [  8,   9, 100,  11],
           [ 12,  13,  14, 100]])

    Now, we create indices to manipulate a 3-D array:

    >>> d3 = np.diag_indices(2, 3)
    >>> d3
    (array([0, 1]), array([0, 1]), array([0, 1]))

    And use it to set the diagonal of an array of zeros to 1:

    >>> a = np.zeros((2, 2, 2), dtype=int)
    >>> a[d3] = 1
    >>> a
    array([[[1, 0],
            [0, 0]],
           [[0, 0],
            [0, 1]]])

    """

    if not use_origin_backend():
        return dpnp_diag_indices(n, ndim)

    return call_origin(numpy.diag_indices, n, ndim)


def diag_indices_from(x1):
    """
    Return the indices to access the main diagonal of an n-dimensional array.

    For full documentation refer to :obj:`numpy.diag_indices_from`.

    See also
    --------
    :obj:`diag_indices` : Return the indices to access the main
                          diagonal of an array.

    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        # original limitation
        if not x1_desc.ndim >= 2:
            pass

        # original limitation
        # For more than d=2, the strided formula is only valid for arrays with
        # all dimensions equal, so we check first.
        elif not numpy.alltrue(numpy.diff(x1_desc.shape) == 0):  # TODO: replace alltrue and diff funcs with dpnp own ones
            pass
        else:
            return dpnp_diag_indices(x1_desc.shape[0], x1_desc.ndim)

    return call_origin(numpy.diag_indices_from, x1)


def diagonal(x1, offset=0, axis1=0, axis2=1):
    """
    Return specified diagonals.

    For full documentation refer to :obj:`numpy.diagonal`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Parameters ``axis1`` and ``axis2`` are supported only with default values.
    Otherwise the function will be executed sequentially on CPU.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        if not isinstance(offset, int):
            pass
        elif offset < 0:
            pass
        elif axis1 != 0:
            pass
        elif axis2 != 1:
            pass
        else:
            return dpnp_diagonal(x1_desc, offset).get_pyobj()

    return call_origin(numpy.diagonal, x1, offset, axis1, axis2)


def fill_diagonal(x1, val, wrap=False):
    """
    Fill the main diagonal of the given array of any dimensionality.

    For full documentation refer to :obj:`numpy.fill_diagonal`.

    Limitations
    -----------
    Parameter ``wrap`` is supported only with default values.

    See Also
    --------
    :obj:`dpnp.diag_indices` : Return the indices to access the main diagonal of an array.
    :obj:`dpnp.diag_indices_from` : Return the indices to access the main diagonal of an n-dimensional array.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1, copy_when_strides=False)
    if x1_desc:
        if not dpnp.isscalar(val):
            pass
        elif wrap:
            pass
        else:
            return dpnp_fill_diagonal(x1_desc, val)

    return call_origin(numpy.fill_diagonal, x1, val, wrap, dpnp_inplace=True)


def indices(dimensions, dtype=int, sparse=False):
    """
    Return an array representing the indices of a grid.

    For full documentation refer to :obj:`numpy.indices`.

    Limitations
    -----------
    Parameters ``dtype`` and ``sparse`` are supported only with default values.
    Parameter ``dimensions`` is supported with len <=2.
    """

    if not isinstance(dimensions, (tuple, list)):
        pass
    elif len(dimensions) > 2 or len(dimensions) == 0:
        pass
    elif dtype != int:
        pass
    elif sparse:
        pass
    else:
        return dpnp_indices(dimensions)

    return call_origin(numpy.indices, dimensions, dtype, sparse)


def nonzero(x1):
    """
    Return the indices of the elements that are non-zero.

    For full documentation refer to :obj:`numpy.nonzero`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Otherwise the function will be executed sequentially on CPU.
    Input array data types are limited by supported DPNP :ref:`Data types`.

    See Also
    --------
    :obj:`dpnp.flatnonzero` : Return indices that are non-zero in
                              the flattened version of the input array.
    :obj:`dpnp.count_nonzero` : Counts the number of non-zero elements
                                in the input array.

    Notes
    -----
    While the nonzero values can be obtained with ``a[nonzero(a)]``, it is
    recommended to use ``x[x.astype(bool)]`` or ``x[x != 0]`` instead, which
    will correctly handle 0-d arrays.

    Examples
    --------
    >>> import dpnp as np
    >>> x = np.array([[3, 0, 0], [0, 4, 0], [5, 6, 0]])
    >>> out = np.nonzero(x)
    >>> for arr in out:
    >>>     [i for i in arr]
    [0, 1, 2, 2]
    [0, 1, 0, 1]

    >>> x2 = np.array([3, 0, 0, 0, 4, 0, 5, 6, 0])
    >>> out2 = np.nonzero(x2)
    >>> for arr in out2:
    >>>     [i for i in arr]
    [0, 4, 6, 7]

    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        return dpnp_nonzero(x1_desc)

    return call_origin(numpy.nonzero, x1)


def place(x1, mask, vals):
    """
    Change elements of an array based on conditional and input values.
    For full documentation refer to :obj:`numpy.place`.

    Limitations
    -----------
    Input arrays ``arr`` and ``mask``  are supported as :obj:`dpnp.ndarray`.
    Parameter ``vals`` is supported as 1-D sequence.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    mask_desc = dpnp.get_dpnp_descriptor(mask)
    vals_desc = dpnp.get_dpnp_descriptor(vals)
    if x1_desc and mask_desc and vals_desc:
        return dpnp_place(x1_desc, mask, vals_desc)

    return call_origin(numpy.place, x1, mask, vals, dpnp_inplace=True)


def put(x1, ind, v, mode='raise'):
    """
    Replaces specified elements of an array with given values.
    For full documentation refer to :obj:`numpy.put`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Not supported parameter mode.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1, copy_when_strides=False)
    if x1_desc:
        if mode != 'raise':
            pass
        elif type(ind) != type(v):
            pass
        elif numpy.max(ind) >= x1_desc.size or numpy.min(ind) + x1_desc.size < 0:
            pass
        else:
            return dpnp_put(x1_desc, ind, v)

    return call_origin(numpy.put, x1, ind, v, mode, dpnp_inplace=True)


def put_along_axis(x1, indices, values, axis):
    """
    Put values into the destination array by matching 1d index and data slices.
    For full documentation refer to :obj:`numpy.put_along_axis`.

    See Also
    --------
    :obj:`take_along_axis` : Take values from the input array by matching 1d index and data slices.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    indices_desc = dpnp.get_dpnp_descriptor(indices)
    values_desc = dpnp.get_dpnp_descriptor(values)
    if x1_desc and indices_desc and values_desc:
        if x1_desc.ndim != indices_desc.ndim:
            pass
        elif not isinstance(axis, int):
            pass
        elif axis >= x1_desc.ndim:
            pass
        elif indices_desc.size != values_desc.size:
            pass
        else:
            return dpnp_put_along_axis(x1_desc, indices_desc, values_desc, axis)

    return call_origin(numpy.put_along_axis, x1, indices, values, axis, dpnp_inplace=True)


def putmask(x1, mask, values):
    """
    Changes elements of an array based on conditional and input values.
    For full documentation refer to :obj:`numpy.putmask`.

    Limitations
    -----------
    Input arrays ``arr``, ``mask`` and ``values``  are supported as :obj:`dpnp.ndarray`.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1, copy_when_strides=False)
    mask_desc = dpnp.get_dpnp_descriptor(mask)
    values_desc = dpnp.get_dpnp_descriptor(values)
    if x1_desc and mask_desc and values_desc:
        return dpnp_putmask(x1_desc, mask_desc, values_desc)

    return call_origin(numpy.putmask, x1, mask, values, dpnp_inplace=True)


def select(condlist, choicelist, default=0):
    """
    Return an array drawn from elements in choicelist, depending on conditions.
    For full documentation refer to :obj:`numpy.select`.

    Limitations
    -----------
    Arrays of input lists are supported as :obj:`dpnp.ndarray`.
    Parameter ``default`` are supported only with default values.
    """

    if not use_origin_backend():
        if not isinstance(condlist, list):
            pass
        elif not isinstance(choicelist, list):
            pass
        elif len(condlist) != len(choicelist):
            pass
        else:
            val = True
            size_ = condlist[0].size
            for i in range(len(condlist)):
                if condlist[i].size != size_ or choicelist[i].size != size_:
                    val = False
            if not val:
                pass
            else:
                return dpnp_select(condlist, choicelist, default).get_pyobj()

    return call_origin(numpy.select, condlist, choicelist, default)


def take(x1, indices, axis=None, out=None, mode='raise'):
    """
    Take elements from an array.
    For full documentation refer to :obj:`numpy.take`.

    Limitations
    -----------
    Input array is supported as :obj:`dpnp.ndarray`.
    Parameters ``axis``, ``out`` and ``mode`` are supported only with default values.
    Parameter ``indices`` is supported as :obj:`dpnp.ndarray`.

    See Also
    --------
    :obj:`dpnp.compress` : Take elements using a boolean mask.
    :obj:`take_along_axis` : Take elements by matching the array and the index arrays.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    indices_desc = dpnp.get_dpnp_descriptor(indices)
    if x1_desc and indices_desc:
        if axis is not None:
            pass
        elif out is not None:
            pass
        elif mode != 'raise':
            pass
        else:
            return dpnp_take(x1_desc, indices_desc).get_pyobj()

    return call_origin(numpy.take, x1, indices, axis, out, mode)


def take_along_axis(x1, indices, axis):
    """
    Take values from the input array by matching 1d index and data slices.
    For full documentation refer to :obj:`numpy.take_along_axis`.

    See Also
    --------
    :obj:`dpnp.take` : Take along an axis, using the same indices for every 1d slice.
    :obj:`put_along_axis` : Put values into the destination array by matching 1d index and data slices.
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    indices_desc = dpnp.get_dpnp_descriptor(indices)
    if x1_desc and indices_desc:
        if x1_desc.ndim != indices_desc.ndim:
            pass
        elif not isinstance(axis, int):
            pass
        elif axis >= x1_desc.ndim:
            pass
        elif x1_desc.ndim == indices_desc.ndim:
            val_list = []
            for i in list(indices_desc.shape)[:-1]:
                if i == 1:
                    val_list.append(True)
                else:
                    val_list.append(False)
            if not all(val_list):
                pass
            else:
                return dpnp_take_along_axis(x1, indices, axis)
        else:
            return dpnp_take_along_axis(x1, indices, axis)

    return call_origin(numpy.take_along_axis, x1, indices, axis)


def tril_indices(n, k=0, m=None):
    """
    Return the indices for the lower-triangle of an (n, m) array.

    Parameters
    ----------
    n : int
        The row dimension of the arrays for which the returned
        indices will be valid.

    k : int, optional
        Diagonal offset (see `tril` for details).

    m : int, optional
        The column dimension of the arrays for which the returned
        arrays will be valid.
        By default `m` is taken equal to `n`.

    Returns
    -------
    inds : tuple of arrays
        The indices for the triangle. The returned tuple contains two arrays,
        each with the indices along one dimension of the array.

    """

    if not use_origin_backend():
        if isinstance(n, int) and isinstance(k, int) \
                and (isinstance(m, int) or m is None):
            return dpnp_tril_indices(n, k, m)

    return call_origin(numpy.tril_indices, n, k, m)


def tril_indices_from(x1, k=0):
    """
    Return the indices for the lower-triangle of arr.
    See `tril_indices` for full details.

    Parameters
    ----------
    arr : array_like
        The indices will be valid for square arrays whose dimensions are
        the same as arr.

    k : int, optional
        Diagonal offset (see `tril` for details).
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        if isinstance(k, int):
            return dpnp_tril_indices_from(x1_desc, k)

    return call_origin(numpy.tril_indices_from, x1, k)


def triu_indices(n, k=0, m=None):
    """
    Return the indices for the upper-triangle of an (n, m) array.

    Parameters
    ----------
    n : int
        The size of the arrays for which the returned indices will
        be valid.

    k : int, optional
        Diagonal offset (see `triu` for details).

    m : int, optional
        The column dimension of the arrays for which the returned
        arrays will be valid.
        By default `m` is taken equal to `n`.

    Returns
    -------
    inds : tuple, shape(2) of ndarrays, shape(`n`)
        The indices for the triangle. The returned tuple contains two arrays,
        each with the indices along one dimension of the array.  Can be used
        to slice a ndarray of shape(`n`, `n`).
    """

    if not use_origin_backend():
        if isinstance(n, int) and isinstance(k, int) \
                and (isinstance(m, int) or m is None):
            return dpnp_triu_indices(n, k, m)

    return call_origin(numpy.triu_indices, n, k, m)


def triu_indices_from(x1, k=0):
    """
    Return the indices for the lower-triangle of arr.
    See `tril_indices` for full details.

    Parameters
    ----------
    arr : array_like
        The indices will be valid for square arrays whose dimensions are
        the same as arr.

    k : int, optional
        Diagonal offset (see `tril` for details).
    """

    x1_desc = dpnp.get_dpnp_descriptor(x1)
    if x1_desc:
        if isinstance(k, int):
            return dpnp_triu_indices_from(x1_desc, k)

    return call_origin(numpy.triu_indices_from, x1, k)



import pytest
import unittest

import dpnp.random
import numpy
from numpy.testing import (assert_allclose, assert_array_equal)
import math


class TestDistribution(unittest.TestCase):

    def check_extreme_value(self, dist_name, val, params):
        seed = 28041990
        size = 10
        dpnp.random.seed(seed)
        res = dpnp.asnumpy(getattr(dpnp.random, dist_name)(size=size, **params))
        assert len(numpy.unique(res)) == 1
        assert numpy.unique(res)[0] == val

    def check_moments(self, dist_name, expected_mean, expected_var, params, size=10**5):
        seed = 28041995
        dpnp.random.seed(seed)
        res = dpnp.asnumpy(getattr(dpnp.random, dist_name)(size=size, **params))
        var = numpy.var(res)
        mean = numpy.mean(res)
        assert math.isclose(var, expected_var, abs_tol=0.1)
        assert math.isclose(mean, expected_mean, abs_tol=0.1)

    def check_invalid_args(self, dist_name, params):
        size = 10
        with pytest.raises(ValueError):
            getattr(dpnp.random, dist_name)(size=size, **params)

    def check_seed(self, dist_name, params):
        seed = 28041990
        size = 10
        dpnp.random.seed(seed)
        a1 = dpnp.asnumpy(getattr(dpnp.random, dist_name)(size=size, **params))
        dpnp.random.seed(seed)
        a2 = dpnp.asnumpy(getattr(dpnp.random, dist_name)(size=size, **params))
        assert_allclose(a1, a2, rtol=1e-07, atol=0)


@pytest.mark.parametrize("func",
                         [dpnp.random.chisquare,
                          dpnp.random.rand,
                          dpnp.random.randn],
                         ids=['chisquare', 'rand', 'randn'])
def test_input_size(func):
    output_shape = (10,)
    size = 10
    df = 3  # for dpnp.random.chisquare
    if func == dpnp.random.chisquare:
        res = func(df, size)
    else:
        res = func(size)
    assert output_shape == res.shape


@pytest.mark.parametrize("func",
                         [dpnp.random.chisquare,
                          dpnp.random.random,
                          dpnp.random.random_sample,
                          dpnp.random.ranf,
                          dpnp.random.sample],
                         ids=['chisquare', 'random', 'random_sample',
                              'ranf', 'sample'])
def test_input_shape(func):
    shape = (10, 5)
    df = 3  # for dpnp.random.chisquare
    if func == dpnp.random.chisquare:
        res = func(df, shape)
    else:
        res = func(shape)
    assert shape == res.shape


@pytest.mark.parametrize("func",
                         [dpnp.random.random,
                          dpnp.random.random_sample,
                          dpnp.random.ranf,
                          dpnp.random.sample,
                          dpnp.random.rand],
                         ids=['random', 'random_sample',
                              'ranf', 'sample',
                              'rand'])
def test_check_otput(func):
    shape = (10, 5)
    size = 10 * 5
    if func == dpnp.random.rand:
        res = func(size)
    else:
        res = func(shape)
#    assert numpy.all(res >= 0)
#    assert numpy.all(res < 1)
    res_as_numpy = dpnp.asnumpy(res)
    for i in range(res_as_numpy.size):
        assert res_as_numpy[i] >= 0.0
        assert res_as_numpy[i] < 1.0


@pytest.mark.parametrize("func",
                         [dpnp.random.random,
                          dpnp.random.random_sample,
                          dpnp.random.ranf,
                          dpnp.random.sample,
                          dpnp.random.rand],
                         ids=['random', 'random_sample',
                              'ranf', 'sample', 'rand'])
def test_seed(func):
    seed = 28041990
    size = 100
    shape = (100, 1)
    if func in [dpnp.random.rand, dpnp.random.randn]:
        args = size
    else:
        args = shape
    dpnp.random.seed(seed)
    a1 = func(args)
    dpnp.random.seed(seed)
    a2 = func(args)
    assert_allclose(a1, a2, rtol=1e-07, atol=0)


def test_randn_normal_distribution():
    """
    Check the moments of the normal distribution sample obtained
    from ``dpnp.random.randn``.

    """

    seed = 28041995
    pts = 10**5
    alpha = 0.05
    expected_mean = 0.0
    expected_var = 1.0

    dpnp.random.seed(seed)
    res = dpnp.asnumpy(dpnp.random.randn(pts))
    var = numpy.var(res)
    mean = numpy.mean(res)
    assert math.isclose(var, expected_var, abs_tol=0.03)
    assert math.isclose(mean, expected_mean, abs_tol=0.03)


class TestDistributionsBeta(TestDistribution):

    def test_moments(self):
        a = 2.56
        b = 0.8
        expected_mean = a / (a + b)
        expected_var = (a * b) / ((a + b)**2 * (a + b + 1))
        self.check_moments('beta', expected_mean, expected_var, {'a': a, 'b': b})

    def test_invalid_args(self):
        a = 3.0   # OK
        b = -1.0  # positive `b` is expected
        self.check_invalid_args('beta', {'a': a, 'b': b})
        a = -1.0  # positive `a` is expected
        b = 3.0   # OK
        self.check_invalid_args('beta', {'a': a, 'b': b})

    def test_seed(self):
        a = 2.56
        b = 0.8
        self.check_seed('beta', {'a': a, 'b': b})


class TestDistributionsBinomial(TestDistribution):

    def test_extreme_value(self):
        n = 5
        p = 0.0
        expected_val = p
        self.check_extreme_value('binomial', expected_val, {'n': n, 'p': p})
        n = 0
        p = 0.5
        expected_val = n
        self.check_extreme_value('binomial', expected_val, {'n': n, 'p': p})
        n = 5
        p = 1.0
        expected_val = n
        self.check_extreme_value('binomial', expected_val, {'n': n, 'p': p})

    def test_moments(self):
        n = 5
        p = 0.8
        expected_mean = n * p
        expected_var = n * p * (1 - p)
        self.check_moments('binomial', expected_mean,
                           expected_var, {'n': n, 'p': p})

    def test_invalid_args(self):
        n = -5     # non-negative `n` is expected
        p = 0.4    # OK
        self.check_invalid_args('binomial', {'n': n, 'p': p})
        n = 5      # OK
        p = -0.5   # `p` is expected from [0, 1]
        self.check_invalid_args('binomial', {'n': n, 'p': p})

    def test_seed(self):
        n, p = 10, .5  # number of trials, probability of each trial
        self.check_seed('binomial', {'n': n, 'p': p})


class TestDistributionsChisquare(TestDistribution):

    def test_invalid_args(self):
        df = -1  # positive `df` is expected
        self.check_invalid_args('chisquare', {'df': df})

    def test_seed(self):
        df = 3  # number of degrees of freedom
        self.check_seed('chisquare', {'df': df})


class TestDistributionsExponential(TestDistribution):

    def test_invalid_args(self):
        scale = -1  # non-negative `scale` is expected
        self.check_invalid_args('exponential', {'scale': scale})

    def test_seed(self):
        scale = 3  # number of degrees of freedom
        self.check_seed('exponential', {'scale': scale})


class TestDistributionsF(TestDistribution):

    def test_moments(self):
        dfnum = 12.56
        dfden = 13.0
        # for dfden > 2
        expected_mean = dfden / (dfden - 2)
        # for dfden > 4
        expected_var = 2 * (dfden ** 2) * (dfnum + dfden - 2) / (dfnum * ((dfden - 2) ** 2) * ((dfden - 4)))
        self.check_moments('f', expected_mean, expected_var,
                           {'dfnum': dfnum, 'dfden': dfden})

    def test_invalid_args(self):
        size = 10
        dfnum = -1.0   # positive `dfnum` is expected
        dfden = 1.0    # OK
        self.check_invalid_args('f', {'dfnum': dfnum, 'dfden': dfden})
        dfnum = 1.0    # OK
        dfden = -1.0   # positive `dfden` is expected
        self.check_invalid_args('f', {'dfnum': dfnum, 'dfden': dfden})

    def test_seed(self):
        dfnum = 3.56   # `dfden` param for Wald distr
        dfden = 2.8    # `dfden` param for Wald distr
        self.check_seed('f', {'dfnum': dfnum, 'dfden': dfden})


class TestDistributionsGamma(TestDistribution):

    def test_moments(self):
        shape = 2.56
        scale = 0.8
        expected_mean = shape * scale
        expected_var = shape * scale * scale
        self.check_moments('gamma', expected_mean, expected_var,
                           {'shape': shape, 'scale': scale})

    def test_invalid_args(self):
        size = 10
        shape = -1   # non-negative `shape` is expected
        self.check_invalid_args('gamma', {'shape': shape})
        shape = 1.0   # OK
        scale = -1.0  # non-negative `shape` is expected
        self.check_invalid_args('gamma', {'shape': shape, 'scale': scale})

    def test_seed(self):
        shape = 3.0  # shape param for gamma distr
        self.check_seed('gamma', {'shape': shape})


class TestDistributionsGeometric(TestDistribution):

    def test_extreme_value(self):
        p = 1.0
        expected_val = p
        self.check_extreme_value('geometric', expected_val, {'p': p})

    def test_moments(self):
        p = 0.8
        expected_mean = (1 - p) / p
        expected_var = (1 - p) / (p**2)
        self.check_moments('geometric', expected_mean, expected_var, {'p': p})

    def test_invalid_args(self):
        size = 10
        p = -1.0  # `p` is expected from (0, 1]
        self.check_invalid_args('geometric', {'p': p})

    def test_seed(self):
        p = 0.8
        self.check_seed('geometric', {'p': p})


class TestDistributionsGumbel(TestDistribution):

    def test_extreme_value(self):
        loc = 5
        scale = 0.0
        expected_val = loc
        self.check_extreme_value('gumbel', expected_val,
                                 {'loc': loc, 'scale': scale})

    def test_moments(self):
        loc = 12
        scale = 0.8
        expected_mean = loc + scale * numpy.euler_gamma
        expected_var = (numpy.pi**2 / 6) * (scale ** 2)
        self.check_moments('gumbel', expected_mean,
                           expected_var, {'loc': loc, 'scale': scale})

    def test_invalid_args(self):
        size = 10
        loc = 3.0     # OK
        scale = -1.0  # non-negative `scale` is expected
        self.check_invalid_args('gumbel', {'loc': loc, 'scale': scale})

    def test_seed(self):
        loc = 2.56
        scale = 0.8
        self.check_seed('gumbel', {'loc': loc, 'scale': scale})


class TestDistributionsHypergeometric(TestDistribution):

    def test_extreme_value(self):
        ngood = 100
        nbad = 0
        nsample = 10
        expected_val = nsample
        self.check_extreme_value('hypergeometric', expected_val,
                                 {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})
        ngood = 0
        nbad = 11
        nsample = 10
        expected_val = 0
        self.check_extreme_value('hypergeometric', expected_val,
                                 {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

    def test_moments(self):
        ngood = 100
        nbad = 2
        nsample = 10
        expected_mean = nsample * (ngood / (ngood + nbad))
        expected_var = expected_mean * (nbad / (ngood + nbad)) * (((ngood + nbad) - nsample) / ((ngood + nbad) - 1))
        self.check_moments('hypergeometric', expected_mean, expected_var,
                           {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

    def test_invalid_args(self):
        size = 10
        ngood = 100    # OK
        nbad = 2       # OK
        nsample = -10  # non-negative `nsamp` is expected
        self.check_invalid_args('hypergeometric',
                                {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

        ngood = 100    # OK
        nbad = -2      # non-negative `nbad` is expected
        nsample = 10   # OK
        self.check_invalid_args('hypergeometric',
                                {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

        ngood = -100   # non-negative `ngood` is expected
        nbad = 2       # OK
        nsample = 10   # OK
        self.check_invalid_args('hypergeometric',
                                {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

        ngood = 10
        nbad = 2
        nsample = 100
        # ngood + nbad >= nsample expected
        self.check_invalid_args('hypergeometric',
                                {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

        ngood = 10   # OK
        nbad = 2     # OK
        nsample = 0  # `nsample` is expected > 0
        self.check_invalid_args('hypergeometric',
                                {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})

    def test_seed(self):
        ngood = 100
        nbad = 2
        nsample = 10
        self.check_seed('hypergeometric',
                        {'ngood': ngood, 'nbad': nbad, 'nsample': nsample})


class TestDistributionsLaplace(TestDistribution):

    def test_extreme_value(self):
        loc = 5
        scale = 0.0
        expected_val = scale
        self.check_extreme_value('laplace', expected_val,
                                 {'loc': loc, 'scale': scale})

    def test_moments(self):
        loc = 2.56
        scale = 0.8
        expected_mean = loc
        expected_var = 2 * scale * scale
        self.check_moments('laplace', expected_mean,
                           expected_var, {'loc': loc, 'scale': scale})

    def test_invalid_args(self):
        loc = 3.0     # OK
        scale = -1.0  # positive `b` is expected
        self.check_invalid_args('laplace',
                                {'loc': loc, 'scale': scale})

    def test_seed(self):
        loc = 2.56
        scale = 0.8
        self.check_seed('laplace', {'loc': loc, 'scale': scale})


class TestDistributionsLogistic(TestDistribution):

    def test_moments(self):
        loc = 2.56
        scale = 0.8
        expected_mean = loc
        expected_var = (scale ** 2) * (numpy.pi ** 2) / 3
        self.check_moments('logistic', expected_mean,
                           expected_var, {'loc': loc, 'scale': scale})

    def test_invalid_args(self):
        loc = 3.0     # OK
        scale = -1.0  # non-negative `scale` is expected
        self.check_invalid_args('logistic',
                                {'loc': loc, 'scale': scale})

    def test_seed(self):
        loc = 2.56
        scale = 0.8
        self.check_seed('logistic', {'loc': loc, 'scale': scale})


class TestDistributionsLognormal(TestDistribution):

    def test_extreme_value(self):
        mean = 0.5
        sigma = 0.0
        expected_val = numpy.exp(mean + (sigma ** 2) / 2)
        self.check_extreme_value('lognormal', expected_val,
                                 {'mean': mean, 'sigma': sigma})

    def test_moments(self):
        mean = 0.5
        sigma = 0.8
        expected_mean = numpy.exp(mean + (sigma ** 2) / 2)
        expected_var = (numpy.exp(sigma**2) - 1) * numpy.exp(2 * mean + sigma**2)
        self.check_moments('lognormal', expected_mean,
                           expected_var, {'mean': mean, 'sigma': sigma})

    def test_invalid_args(self):
        mean = 0.0
        sigma = -1.0  # non-negative `sigma` is expected
        self.check_invalid_args('lognormal', {'mean': mean, 'sigma': sigma})

    def test_seed(self):
        mean = 0.0
        sigma = 0.8
        self.check_seed('lognormal', {'mean': mean, 'sigma': sigma})


class TestDistributionsMultinomial(TestDistribution):

    def test_extreme_value(self):
        n = 0
        pvals = [1 / 6.] * 6
        self.check_extreme_value('multinomial', n, {'n': n, 'pvals': pvals})

    def test_moments(self):
        n = 10
        pvals = [1 / 6.] * 6
        size = 10**5
        expected_mean = n * pvals[0]
        expected_var = n * pvals[0] * (1 - pvals[0])
        self.check_moments('multinomial', expected_mean,
                           expected_var, {'n': n, 'pvals': pvals})

    def test_check_sum(self):
        seed = 28041990
        size = 1
        n = 20
        pvals = [1 / 6.] * 6
        dpnp.random.seed(seed)
        res = dpnp.random.multinomial(n, pvals, size)
        assert_allclose(n, dpnp.asnumpy(res).sum(), rtol=1e-07, atol=0)

    def test_invalid_args(self):
        n = -10                # parameter `n`, non-negative expected
        pvals = [1 / 6.] * 6   # parameter `pvals`, OK
        self.check_invalid_args('multinomial', {'n': n, 'pvals': pvals})
        n = 10                 # parameter `n`, OK
        pvals = [-1 / 6.] * 6  # parameter `pvals`, sum(pvals) expected between [0, 1]
        self.check_invalid_args('multinomial', {'n': n, 'pvals': pvals})

    def test_seed(self):
        n = 20
        pvals = [1 / 6.] * 6
        self.check_seed('multinomial', {'n': n, 'pvals': pvals})


    def test_seed1(self):
        # pvals_size >= ntrial * 16 && ntrial <= 16
        n = 4
        pvals_size = 16 * n
        pvals = [1 / pvals_size] * pvals_size
        self.check_seed('multinomial', {'n': n, 'pvals': pvals})


class TestDistributionsMultivariateNormal(TestDistribution):

    def test_moments(self):
        seed = 2804183
        dpnp.random.seed(seed)
        mean = [2.56, 3.23]
        cov = [[1, 0], [0, 1]]
        size = 10**5
        res = dpnp.asnumpy(dpnp.random.multivariate_normal(mean=mean, cov=cov, size=size))
        res_mean = [numpy.mean(res.T[0]), numpy.mean(res.T[1])]
        assert_allclose(res_mean, mean, rtol=1e-02, atol=0)

    def test_invalid_args(self):
        mean = [2.56, 3.23]  # OK
        cov = [[1, 0]]       # `mean` and `cov` must have same length
        self.check_invalid_args('multivariate_normal', {'mean': mean, 'cov': cov})
        mean = [[2.56, 3.23]]   # `mean` must be 1 dimensional
        cov = [[1, 0], [0, 1]]  # OK
        self.check_invalid_args('multivariate_normal', {'mean': mean, 'cov': cov})
        mean = [2.56, 3.23]  # OK
        cov = [1, 0, 0, 1]   # `cov` must be 2 dimensional and square
        self.check_invalid_args('multivariate_normal', {'mean': mean, 'cov': cov})

    def test_output_shape_check(self):
        seed = 28041990
        size = 100
        mean = [2.56, 3.23]
        cov = [[1, 0], [0, 1]]
        expected_shape = (100, 2)
        dpnp.random.seed(seed)
        res = dpnp.random.multivariate_normal(mean, cov, size=100)
        assert res.shape == expected_shape

    def test_seed(self):
        mean = [2.56, 3.23]
        cov = [[1, 0], [0, 1]]
        self.check_seed('multivariate_normal', {'mean': mean, 'cov': cov})


class TestDistributionsNegativeBinomial(TestDistribution):

    def test_extreme_value(self):
        seed = 28041990
        dpnp.random.seed(seed)
        n = 5
        p = 1.0
        check_val = 0.0
        self.check_extreme_value('negative_binomial', check_val, {'n': n, 'p': p})
        n = 5
        p = 0.0
        res = dpnp.asnumpy(dpnp.random.negative_binomial(n=n, p=p, size=10))
        check_val = numpy.iinfo(res.dtype).min
        assert len(numpy.unique(res)) == 1
        assert numpy.unique(res)[0] == check_val

    def test_invalid_args(self):
        n = 10    # parameter `n`, OK
        p = -0.5  # parameter `p`, expected between [0, 1]
        self.check_invalid_args('negative_binomial', {'n': n, 'p': p})
        n = -10   # parameter `n`, expected non-negative
        p = 0.5   # parameter `p`, OK
        self.check_invalid_args('negative_binomial', {'n': n, 'p': p})

    def test_seed(self):
        n, p = 10, .5  # number of trials, probability of each trial
        self.check_seed('negative_binomial', {'n': n, 'p': p})


class TestDistributionsNormal(TestDistribution):

    def test_extreme_value(self):
        loc = 5
        scale = 0.0
        expected_val = loc
        self.check_extreme_value('normal', expected_val, {'loc': loc, 'scale': scale})

    def test_moments(self):
        loc = 2.56
        scale = 0.8
        expected_mean = loc
        expected_var = scale**2
        self.check_moments('normal', expected_mean,
                           expected_var, {'loc': loc, 'scale': scale})

    def test_invalid_args(self):
        loc = 3.0     # OK
        scale = -1.0  # non-negative `scale` is expected
        self.check_invalid_args('normal', {'loc': loc, 'scale': scale})

    def test_seed(self):
        loc = 2.56
        scale = 0.8
        self.check_seed('normal', {'loc': loc, 'scale': scale})


class TestDistributionsNoncentralChisquare:

    @pytest.mark.parametrize("df", [5.0, 1.0, 0.5], ids=['df_grt_1', 'df_eq_1', 'df_less_1'])
    def test_moments(self, df):
        nonc = 20.
        expected_mean = df + nonc
        expected_var = 2 * (df + 2 * nonc)
        size = 10**6
        seed = 28041995
        dpnp.random.seed(seed)
        res = dpnp.asnumpy(dpnp.random.noncentral_chisquare(df, nonc, size=size))
        var = numpy.var(res)
        mean = numpy.mean(res)
        assert math.isclose(var, expected_var, abs_tol=0.6)
        assert math.isclose(mean, expected_mean, abs_tol=0.6)

    def test_invalid_args(self):
        size = 10
        df = 5.0     # OK
        nonc = -1.0  # non-negative `nonc` is expected
        with pytest.raises(ValueError):
            dpnp.random.noncentral_chisquare(df, nonc, size=size)
        df = -1.0    # positive `df` is expected
        nonc = 1.0   # OK
        with pytest.raises(ValueError):
            dpnp.random.noncentral_chisquare(df, nonc, size=size)

    @pytest.mark.parametrize("df", [5.0, 1.0, 0.5], ids=['df_grt_1', 'df_eq_1', 'df_less_1'])
    def test_seed(self, df):
        seed = 28041990
        size = 10
        nonc = 1.8
        dpnp.random.seed(seed)
        a1 = dpnp.asarray(dpnp.random.noncentral_chisquare(df, nonc, size=size))
        dpnp.random.seed(seed)
        a2 = dpnp.asarray(dpnp.random.noncentral_chisquare(df, nonc, size=size))
        assert_allclose(a1, a2, rtol=1e-07, atol=0)


class TestDistributionsPareto(TestDistribution):

    def test_moments(self):
        a = 30.0
        expected_mean = a / (a - 1)
        expected_var = a / (((a - 1)**2) * (a - 2))
        self.check_moments('pareto', expected_mean,
                           expected_var, {'a': a})

    def test_invalid_args(self):
        size = 10
        a = -1.0  # positive `a` is expected
        self.check_invalid_args('pareto', {'a': a})

    def test_seed(self):
        a = 3.0  # a param for pareto distr
        self.check_seed('pareto', {'a': a})


class TestDistributionsPoisson(TestDistribution):

    def test_extreme_value(self):
        lam = 0.0
        self.check_extreme_value('poisson', lam, {'lam': lam})

    def test_moments(self):
        lam = 0.8
        expected_mean = lam
        expected_var = lam
        self.check_moments('poisson', expected_mean,
                           expected_var, {'lam': lam})

    def test_invalid_args(self):
        lam = -1.0    # non-negative `lam` is expected
        self.check_invalid_args('poisson', {'lam': lam})

    def test_seed(self):
        lam = 0.8
        self.check_seed('poisson', {'lam': lam})


class TestDistributionsPower(TestDistribution):

    def test_moments(self):
        a = 30.0
        neg_a = -a
        expected_mean = neg_a / (neg_a - 1)
        expected_var = neg_a / (((neg_a - 1)**2) * (neg_a - 2))
        self.check_moments('power', expected_mean,
                           expected_var, {'a': a})

    def test_invalid_args(self):
        size = 10
        a = -1.0  # positive `a` is expected
        self.check_invalid_args('power', {'a': a})

    def test_seed(self):
        a = 3.0  # a param for pareto distr
        self.check_seed('power', {'a': a})


class TestDistributionsRayleigh(TestDistribution):

    def test_extreme_value(self):
        scale = 0.0
        self.check_extreme_value('rayleigh', scale, {'scale': scale})

    def test_moments(self):
        scale = 0.8
        expected_mean = scale * numpy.sqrt(numpy.pi / 2)
        expected_var = ((4 - numpy.pi) / 2) * scale * scale
        self.check_moments('rayleigh', expected_mean,
                           expected_var, {'scale': scale})

    def test_invalid_args(self):
        scale = -1.0  # positive `b` is expected
        self.check_invalid_args('rayleigh', {'scale': scale})

    def test_seed(self):
        scale = 0.8
        self.check_seed('rayleigh', {'scale': scale})


class TestDistributionsStandardCauchy(TestDistribution):

    def test_seed(self):
        self.check_seed('standard_cauchy', {})


class TestDistributionsStandardExponential(TestDistribution):

    def test_moments(self):
        shape = 0.8
        expected_mean = 1.0
        expected_var = 1.0
        self.check_moments('standard_exponential',
                           expected_mean, expected_var, {})

    def test_seed(self):
        self.check_seed('standard_exponential', {})


class TestDistributionsStandardGamma(TestDistribution):

    def test_extreme_value(self):
        self.check_extreme_value('standard_gamma', 0.0, {'shape': 0.0})

    def test_moments(self):
        shape = 0.8
        expected_mean = shape
        expected_var = shape
        self.check_moments('standard_gamma', expected_mean,
                           expected_var, {'shape': shape})

    def test_invalid_args(self):
        shape = -1   # non-negative `shape` is expected
        self.check_invalid_args('standard_gamma', {'shape': shape})

    def test_seed(self):
        self.check_seed('standard_gamma', {'shape': 0.0})


class TestDistributionsStandardNormal(TestDistribution):

    def test_moments(self):
        expected_mean = 0.0
        expected_var = 1.0
        self.check_moments('standard_normal',
                           expected_mean, expected_var, {})

    def test_seed(self):
        self.check_seed('standard_normal', {})


class TestDistributionsStandardT(TestDistribution):

    def test_moments(self):
        df = 300.0
        expected_mean = 0.0
        expected_var = df / (df - 2)
        self.check_moments('standard_t', expected_mean,
                           expected_var, {'df': df})

    def test_invalid_args(self):
        df = 0.0   # positive `df` is expected
        self.check_invalid_args('standard_t', {'df': df})

    def test_seed(self):
        self.check_seed('standard_t', {'df': 10.0})


class TestDistributionsTriangular(TestDistribution):

    def test_moments(self):
        left = 1.0
        mode = 2.0
        right = 3.0
        expected_mean = (left + mode + right) / 3
        expected_var = (left ** 2 + mode ** 2 + right ** 2 - left * mode - left * right - mode * right) / 18
        self.check_moments('triangular', expected_mean,
                           expected_var, {'left': left, 'mode': mode, 'right': right})

    def test_invalid_args(self):
        left = 2.0   # `left` is expected <= `mode`
        mode = 1.0   # `mode` is expected > `left`
        right = 3.0  # OK
        self.check_invalid_args('triangular', {'left': left, 'mode': mode, 'right': right})

        left = 1.0   # OK
        mode = 3.0   # `mode` is expected <= `right`
        right = 2.0  # `right` is expected > `mode`
        self.check_invalid_args('triangular', {'left': left, 'mode': mode, 'right': right})

    def test_seed(self):
        left = 1.0
        mode = 2.0
        right = 3.0
        self.check_seed('triangular', {'left': left, 'mode': mode, 'right': right})


class TestDistributionsUniform(TestDistribution):

    def test_extreme_value(self):
        low = 1.0
        high = 1.0
        expected_val = low
        self.check_extreme_value('uniform', expected_val,
                                 {'low': low, 'high': high})

    def test_moments(self):
        low = 1.0
        high = 2.0
        expected_mean = (low + high) / 2
        expected_var = ((high - low) ** 2) / 12
        self.check_moments('uniform', expected_mean,
                           expected_var, {'low': low, 'high': high})

    def test_seed(self):
        low = 1.0
        high = 2.0
        self.check_seed('uniform', {'low': low, 'high': high})


class TestDistributionsVonmises:

    @pytest.mark.parametrize("kappa", [5.0, 0.5], ids=['large_kappa', 'small_kappa'])
    def test_moments(self, kappa):
        size = 10**6
        mu = 2.

        numpy_res = numpy.random.vonmises(mu, kappa, size=size)
        expected_mean = numpy.mean(numpy_res)
        expected_var = numpy.var(numpy_res)

        res = dpnp.asnumpy(dpnp.random.vonmises(mu, kappa, size=size))
        var = numpy.var(res)
        mean = numpy.mean(res)
        assert math.isclose(var, expected_var, abs_tol=0.6)
        assert math.isclose(mean, expected_mean, abs_tol=0.6)

    def test_invalid_args(self):
        size = 10
        mu = 5.0      # OK
        kappa = -1.0  # non-negative `kappa` is expected
        with pytest.raises(ValueError):
            dpnp.random.vonmises(mu, kappa, size=size)

    @pytest.mark.parametrize("kappa", [5.0, 0.5], ids=['large_kappa', 'small_kappa'])
    def test_seed(self, kappa):
        seed = 28041990
        size = 10
        mu = 2.
        dpnp.random.seed(seed)
        a1 = dpnp.asarray(dpnp.random.vonmises(mu, kappa, size=size))
        dpnp.random.seed(seed)
        a2 = dpnp.asarray(dpnp.random.vonmises(mu, kappa, size=size))
        assert_allclose(a1, a2, rtol=1e-07, atol=0)


class TestDistributionsWald(TestDistribution):

    def test_moments(self):
        size = 5 * 10**6
        mean = 3.56
        scale = 2.8
        expected_mean = mean
        expected_var = (mean ** 3) / scale
        self.check_moments('wald', expected_mean, expected_var,
                           {'mean': mean, 'scale': scale}, size=size)

    def test_invalid_args(self):
        size = 10
        mean = -1.0   # positive `mean` is expected
        scale = 1.0   # OK
        self.check_invalid_args('wald', {'mean': mean, 'scale': scale})
        mean = 1.0    # OK
        scale = -1.0  # positive `scale` is expected
        self.check_invalid_args('wald', {'mean': mean, 'scale': scale})

    def test_seed(self):
        mean = 3.56   # `mean` param for Wald distr
        scale = 2.8   # `scale` param for Wald distr
        self.check_seed('wald', {'mean': mean, 'scale': scale})


class TestDistributionsWeibull(TestDistribution):

    def test_extreme_value(self):
        a = 0.0
        expected_val = a
        self.check_extreme_value('weibull', expected_val, {'a': a})

    def test_invalid_args(self):
        a = -1.0  # non-negative `a` is expected
        self.check_invalid_args('weibull', {'a': a})

    def test_seed(self):
        a = 2.56
        self.check_seed('weibull', {'a': a})


class TestDistributionsZipf(TestDistribution):

    def test_invalid_args(self):
        a = 1.0  # parameter `a` is expected greater than 1.
        self.check_invalid_args('zipf', {'a': a})

    def test_seed(self):
        a = 2.56
        self.check_seed('zipf', {'a': a})


class TestPermutationsTestShuffle:

    @pytest.mark.parametrize("dtype", [dpnp.float32, dpnp.float64, dpnp.int32, dpnp.int64],
                             ids=['float32', 'float64', 'int32', 'int64'])
    def test_shuffle(self, dtype):
        seed = 28041990
        input_x_int64 = dpnp.asarray([1, 2, 3, 4, 5, 6, 7, 8, 9, 0], dtype=dpnp.int64)
        input_x = dpnp.asarray([1, 2, 3, 4, 5, 6, 7, 8, 9, 0], dtype=dtype)
        dpnp.random.seed(seed)
        dpnp.random.shuffle(input_x_int64)  # inplace
        desired_x = dpnp.astype(input_x_int64, dtype)
        dpnp.random.seed(seed)
        dpnp.random.shuffle(input_x)  # inplace
        actual_x = input_x
        assert_array_equal(actual_x, desired_x)

    @pytest.mark.parametrize("dtype", [dpnp.float32, dpnp.float64, dpnp.int32, dpnp.int64],
                             ids=['float32', 'float64', 'int32', 'int64'])
    def test_no_miss_numbers(self, dtype):
        seed = 28041990
        input_x = dpnp.asarray([5, 4, 0, 7, 6, 1, 8, 3, 2, 9], dtype=dtype)
        desired_x = dpnp.sort(input_x)
        dpnp.random.seed(seed)
        dpnp.random.shuffle(input_x)  # inplace
        output_x = input_x
        actual_x = dpnp.sort(output_x)
        assert_array_equal(actual_x, desired_x)

    @pytest.mark.parametrize("conv", [lambda x: dpnp.array([]),
                                      # lambda x: dpnp.astype(dpnp.asarray(x), dpnp.int8),
                                      lambda x: dpnp.astype(dpnp.asarray(x), dpnp.float32),
                                      # lambda x: dpnp.asarray(x).astype(dpnp.complex64),
                                      # lambda x: dpnp.astype(dpnp.asarray(x), object),
                                      lambda x: dpnp.asarray([[i, i] for i in x]),
                                      lambda x: dpnp.vstack([x, x]).T,
                                      lambda x: (dpnp.asarray([(i, i) for i in x], [
                                                              ("a", int), ("b", int)]).view(dpnp.recarray)),
                                      lambda x: dpnp.asarray([(i, i) for i in x],
                                                             [("a", object), ("b", dpnp.int32)])],
                             ids=['lambda x: dpnp.array([])',
                                  # 'lambda x: dpnp.astype(dpnp.asarray(x), dpnp.int8)',
                                  'lambda x: dpnp.astype(dpnp.asarray(x), dpnp.float32)',
                                  # 'lambda x: dpnp.asarray(x).astype(dpnp.complex64)',
                                  # 'lambda x: dpnp.astype(dpnp.asarray(x), object)',
                                  'lambda x: dpnp.asarray([[i, i] for i in x])',
                                  'lambda x: dpnp.vstack([x, x]).T',
                                  'lambda x: (dpnp.asarray([(i, i) for i in x], ['\
                                  '("a", int), ("b", int)]).view(dpnp.recarray))',
                                  'lambda x: dpnp.asarray([(i, i) for i in x], [("a", object), ("b", dpnp.int32)])]'
                                  ]
                             )
    def test_shuffle1(self, conv):
        # `conv` contans test lists, arrays (of various dtypes), and multidimensional
        # versions of both, c-contiguous or not.
        #
        # This test is a modification of the original tests of `numpy.random` (both the same):
        # * tests/test_random.py::TestRandomDist::test_shuffle
        # * tests/test_randomstate.py::TestRandomDist::test_shuffle
        #
        # The original tests do not have a parameterized launch and they
        # do not correctly check of the results for the dpnp RNG engine.

        # Computing desired 1 dim list for given 1 dim list
        # on the current device for the given seed number.
        seed = 1234567890

        dpnp.random.seed(seed)
        list_1d = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
        dpnp_1d = dpnp.array(list_1d)
        dpnp.random.shuffle(dpnp_1d)  # inplace
        dpnp_desired_1d = dpnp_1d
        desired_1d = [i for i in dpnp_desired_1d]

        dpnp.random.seed(seed)
        alist = conv(list_1d)
        dpnp.random.shuffle(alist)  # inplace
        actual = alist
        desired = conv(desired_1d)
        assert_array_equal(actual, desired)

    @pytest.mark.parametrize("conv", [lambda x: x,
                                      lambda x: [(i, i) for i in x]],
                             ids=['lambda x: x',
                                  'lambda x: [(i, i) for i in x]',
                                  ]
                             )
    def test_shuffle1_fallback(self, conv):
        # This is parameterized version of original tests of `numpy.random` (both the same):
        # * tests/test_random.py::TestRandomDist::test_shuffle
        # * tests/test_randomstate.py::TestRandomDist::test_shuffle

        seed = 1234567890

        dpnp.random.seed(seed)
        alist = conv([1, 2, 3, 4, 5, 6, 7, 8, 9, 0])
        dpnp.random.shuffle(alist)
        actual = alist
        desired = conv([0, 1, 9, 6, 2, 4, 5, 8, 7, 3])
        assert_array_equal(actual, desired)
